@@ -1,11 +1,11 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# File              : Ampel-ZTF/ampel/ztf/base/CatalogMatchUnit.py
-# License           : BSD-3-Clause
-# Author            : Jakob van Santen <jakob.van.santen@desy.de>
-# Date              : 10.03.2021
-# Last Modified Date: 10.03.2021
-# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>
+# File:                Ampel-ZTF/ampel/ztf/base/CatalogMatchUnit.py
+# License:             BSD-3-Clause
+# Author:              Jakob van Santen <jakob.van.santen@desy.de>
+# Date:                10.03.2021
+# Last Modified Date:  30.11.2022
+# Last Modified By:    Simeon Reusch <simeon.reusch@desy.de>
 
 from functools import cached_property
 from typing import (
@@ -24,8 +24,8 @@ import backoff
 import requests
 from requests_toolbelt.sessions import BaseUrlSession
 
-from ampel.base.DataUnit import DataUnit
-from ampel.core.AdminUnit import AdminUnit
+from ampel.base.LogicalUnit import LogicalUnit
+from ampel.core.ContextUnit import ContextUnit
 
 
 class BaseConeSearchRequest(TypedDict):
@@ -69,13 +69,13 @@ class BaseConeSearchRequest(TypedDict):
 
 
 class ConeSearchRequest(BaseConeSearchRequest, total=False):
-    keys_to_append: Optional[Sequence[str]]
-    pre_filter: Optional[Dict[str, Any]]
-    post_filter: Optional[Dict[str, Any]]
+    keys_to_append: None | Sequence[str]
+    pre_filter: None | dict[str, Any]
+    post_filter: None | dict[str, Any]
 
 
 class CatalogItem(TypedDict):
-    body: Dict[str, Any]
+    body: dict[str, Any]
     dist_arcsec: float
 
 
@@ -98,7 +98,7 @@ class CatalogMatchUnitBase:
         ra: float,
         dec: float,
         catalogs: Sequence[ConeSearchRequest],
-    ) -> List[bool]:
+    ) -> list[bool]:
         ...
 
     @overload
@@ -108,7 +108,7 @@ class CatalogMatchUnitBase:
         ra: float,
         dec: float,
         catalogs: Sequence[ConeSearchRequest],
-    ) -> List[Optional[CatalogItem]]:
+    ) -> list[None | CatalogItem]:
         ...
 
     @overload
@@ -118,7 +118,7 @@ class CatalogMatchUnitBase:
         ra: float,
         dec: float,
         catalogs: Sequence[ConeSearchRequest],
-    ) -> List[Optional[List[CatalogItem]]]:
+    ) -> list[None | list[CatalogItem]]:
         ...
 
     @backoff.on_exception(
@@ -130,7 +130,8 @@ class CatalogMatchUnitBase:
     @backoff.on_exception(
         backoff.expo,
         requests.HTTPError,
-        giveup=lambda e: e.response.status_code not in {503, 504, 429, 408},
+        giveup=lambda e: not isinstance(e, requests.HTTPError)
+        or e.response.status_code not in {502, 503, 504, 429, 408},
         max_time=60,
     )
     def _cone_search(
@@ -139,9 +140,9 @@ class CatalogMatchUnitBase:
         ra: float,
         dec: float,
         catalogs: Sequence[ConeSearchRequest],
-    ) -> Union[
-        List[bool], List[Optional[CatalogItem]], List[Optional[List[CatalogItem]]]
-    ]:
+    ) -> list[bool] | list[None | CatalogItem] | list[None | list[CatalogItem]]:
+        if not -90 <= dec <= 90:
+            raise ValueError("Declination angle must be within -90 deg <= angle <= 90 deg, got {dec} deg")
         response = self.session.post(
             f"cone_search/{method}",
             json={
@@ -155,23 +156,23 @@ class CatalogMatchUnitBase:
 
     def cone_search_any(
         self, ra: float, dec: float, catalogs: Sequence[ConeSearchRequest]
-    ) -> List[bool]:
+    ) -> list[bool]:
         return self._cone_search("any", ra, dec, catalogs)
 
     def cone_search_nearest(
         self, ra: float, dec: float, catalogs: Sequence[ConeSearchRequest]
-    ) -> List[Optional[CatalogItem]]:
+    ) -> list[None | CatalogItem]:
         return self._cone_search("nearest", ra, dec, catalogs)
 
     def cone_search_all(
         self, ra: float, dec: float, catalogs: Sequence[ConeSearchRequest]
-    ) -> List[Optional[List[CatalogItem]]]:
+    ) -> list[None | list[CatalogItem]]:
         return self._cone_search("all", ra, dec, catalogs)
 
 
-class CatalogMatchUnit(CatalogMatchUnitBase, DataUnit):
+class CatalogMatchUnit(CatalogMatchUnitBase, LogicalUnit):
     """
-    Catalog matching for DataUnits
+    Catalog matching for LogicalUnits
     """
 
     require = ("ampel-ztf/catalogmatch",)
@@ -182,9 +183,9 @@ class CatalogMatchUnit(CatalogMatchUnitBase, DataUnit):
         return BaseUrlSession(base_url=self.resource["ampel-ztf/catalogmatch"])
 
 
-class CatalogMatchAdminUnit(CatalogMatchUnitBase, AdminUnit):
+class CatalogMatchContextUnit(CatalogMatchUnitBase, ContextUnit):
     """
-    Catalog matching for AdminUnits
+    Catalog matching for ContextUnits
     """
 
     @cached_property
