@@ -1,22 +1,22 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# File              : ampel/ztf/t0/alerts/AllConsumingConsumer.py
-# License           : BSD-3-Clause
-# Author            : Jakob van Santen <jakob.van.santen@desy.de>
-# Date              : Unspecified
-# Last Modified Date: 14.11.2018
-# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>
+# File:                ampel/ztf/t0/alerts/AllConsumingConsumer.py
+# License:             BSD-3-Clause
+# Author:              Jakob van Santen <jakob.van.santen@desy.de>
+# Date:                Unspecified
+# Last Modified Date:  14.11.2018
+# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>
 
 import enum
 import json
 import sys
 import time
 import uuid
-from typing import Optional
 
 import confluent_kafka
 
 from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
+from ampel.protocol.LoggerProtocol import LoggerProtocol
 
 
 class KafkaMetrics:
@@ -69,7 +69,9 @@ class KafkaMetrics:
                 for k, v in partition.items():
                     if metric := self._metrics.get(k):
                         # only record value for assigned partitions (i.e. where desired is True)
-                        metric.labels(topic["topic"], partition["partition"]).set(v if partition["desired"] else -1)
+                        metric.labels(topic["topic"], partition["partition"]).set(
+                            v if partition["desired"] else -1
+                        )
 
     def on_consume(self, message):
         kind, ts = message.timestamp()
@@ -92,7 +94,7 @@ class KafkaError(RuntimeError):
     """Picklable wrapper for cimpl.KafkaError"""
 
     def __init__(self, kafka_err):
-        super().__init__(kafka_err.args[0])
+        super().__init__(kafka_err.str())
         self.code = KafkaErrorCode(kafka_err.code())
 
 
@@ -101,7 +103,15 @@ class AllConsumingConsumer:
     Consume messages on all topics beginning with 'ztf_'.
     """
 
-    def __init__(self, broker, timeout=None, topics=["^ztf_.*"], **consumer_config):
+    def __init__(
+        self,
+        broker,
+        timeout=None,
+        topics=["^ztf_.*"],
+        auto_commit=True,
+        logger: None | LoggerProtocol=None,
+        **consumer_config,
+    ):
         """ """
 
         self._metrics = KafkaMetrics.instance()
@@ -109,7 +119,7 @@ class AllConsumingConsumer:
             "bootstrap.servers": broker,
             "default.topic.config": {"auto.offset.reset": "smallest"},
             "enable.auto.commit": True,
-            "receive.message.max.bytes": 2 ** 29,
+            "receive.message.max.bytes": 2**29,
             "auto.commit.interval.ms": 10000,
             "enable.auto.offset.store": False,
             "group.id": uuid.uuid1(),
@@ -121,6 +131,7 @@ class AllConsumingConsumer:
         }
         config.update(**consumer_config)
         self._consumer = confluent_kafka.Consumer(**config)
+        self._logger = logger
 
         self._consumer.subscribe(topics)
         if timeout is None:
@@ -131,13 +142,8 @@ class AllConsumingConsumer:
             self._poll_attempts = max((1, int(timeout / self._poll_interval)))
         self._timeout = timeout
 
-        self._last_message = None
-
-    def __del__(self):
-        # NB: have to explicitly call close() here to prevent
-        # rd_kafka_consumer_close() from segfaulting. See:
-        # https://github.com/confluentinc/confluent-kafka-python/issues/358
-        self._consumer.close()
+        self._offsets: dict[tuple[str,int],int] = {}
+        self._auto_commit = auto_commit
 
     def __next__(self):
         message = self.consume()
@@ -149,17 +155,34 @@ class AllConsumingConsumer:
     def __iter__(self):
         return self
 
-    def consume(self) -> Optional[confluent_kafka.Message]:
+    def commit(self):
+        if self._offsets:
+            offsets = [
+                confluent_kafka.TopicPartition(topic, partition, offset + 1)
+                for (topic, partition), offset in self._offsets.items()
+            ]
+            if self._logger:
+                self._logger.debug(f"Storing offsets: {offsets}")
+            if self._auto_commit:
+                self._consumer.store_offsets(offsets=offsets)
+                self._offsets.clear()
+            else:
+                for toppar in self._consumer.commit(offsets=offsets, asynchronous=False):
+                    if toppar.error:
+                        self._logger.error(f"Commit {toppar} failed with {toppar.error}")
+                    else:
+                        del self._offsets[(toppar.topic, toppar.partition)]
+
+    def consume(self) -> None | confluent_kafka.Message:
         """
         Block until one message has arrived, and return it.
-        
+
         Messages returned to the caller marked for committal
         upon the _next_ call to consume().
         """
         # mark the last emitted message for committal
-        if self._last_message is not None:
-            self._consumer.store_offsets(self._last_message)
-        self._last_message = None
+        if self._auto_commit:
+            self.commit()
 
         message = None
         for _ in range(self._poll_attempts):
@@ -175,14 +198,16 @@ class AllConsumingConsumer:
                         confluent_kafka.KafkaError._MAX_POLL_EXCEEDED,
                     ):
                         # bail on timeouts
+                        if self._logger:
+                            self._logger.debug(f"Got {err}")
                         return None
                 break
-        else:
-            return message
 
-        if message.error():
+        if message is None:
+            return message
+        elif message.error():
             raise KafkaError(message.error())
         else:
-            self._last_message = message
+            self._offsets[(message.topic(), message.partition())] = message.offset()
             self._metrics.on_consume(message)
             return message
