@@ -1,12 +1,12 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# File              : Ampel-ZTF/ampel/ztf/t3/skyportal/SkyPortalClient.py
-# Author            : Jakob van Santen <jakob.van.santen@desy.de>
-# Date              : 16.09.2020
-# Last Modified Date: 16.09.2020
-# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>
+# File:                Ampel-ZTF/ampel/ztf/t3/skyportal/SkyPortalClient.py
+# Author:              Jakob van Santen <jakob.van.santen@desy.de>
+# Date:                16.09.2020
+# Last Modified Date:  24.11.2022
+# Last Modified By:    Simeon Reusch <simeon.reusch@desy.de>
 
-import asyncio, base64, gzip, io, json, math, time, sys, aiohttp, backoff, pydantic.errors
+import asyncio, base64, gzip, io, json, math, time, aiohttp, backoff
 import numpy as np
 
 from collections import defaultdict
@@ -15,45 +15,25 @@ from datetime import datetime
 from astropy.io import fits
 from matplotlib.colors import Normalize
 from matplotlib.figure import Figure
-from pydantic import AnyHttpUrl
+from typing import Any, TypedDict, overload, TYPE_CHECKING
+from collections.abc import Sequence, Generator, Iterable
+from urllib.parse import urlparse
 
-if sys.version_info.minor > 8:
-    from typing import TypedDict
-else:
-    from typing_extensions import TypedDict
-from typing import (
-    Any,
-    Dict,
-    Generator,
-    Iterable,
-    List,
-    Optional,
-    Tuple,
-    overload,
-    Sequence,
-    Set,
-    TYPE_CHECKING,
-    Union,
-)
-
-from ampel.base.AmpelBaseModel import AmpelBaseModel
+from ampel.base.AmpelUnit import AmpelUnit
 from ampel.log.AmpelLogger import AmpelLogger
 from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
-from ampel.model.Secret import Secret
+from ampel.secret.NamedSecret import NamedSecret
 from ampel.protocol.LoggerProtocol import LoggerProtocol
-from ampel.enum.T2RunState import T2RunState
+from ampel.enum.DocumentCode import DocumentCode
+from ampel.types import Traceless
 from ampel.util.collections import ampel_iter
 from ampel.util.mappings import flatten_dict
 
 if TYPE_CHECKING:
-    from pydantic import AnyUrl
-    from pydantic.fields import ModelField
-    from pydantic.main import BaseConfig
-
     from ampel.config.AmpelConfig import AmpelConfig
     from ampel.content.DataPoint import DataPoint
-    from ampel.content.T2Document import T2Document
     from ampel.view.TransientView import TransientView
+    from ampel.view.T2DocView import T2DocView
 
 
 stat_http_errors = AmpelMetricsRegistry.counter(
@@ -95,13 +75,14 @@ def sanitize_json(obj):
         return obj
 
 
-def encode_t2_body(t2: "T2Document") -> str:
-    assert t2["body"] is not None
-    doc = t2["body"][-1]
+def encode_t2_body(t2: "T2DocView") -> str:
+    assert t2.body is not None
+    doc = t2.body[-1]
+    assert isinstance(doc, dict)
     return base64.b64encode(
         json.dumps(
             {
-                "timestamp": datetime.fromtimestamp(doc["ts"]).isoformat(),
+                "timestamp": datetime.fromtimestamp(t2.meta[-1]["ts"]).isoformat(),
                 **{k: sanitize_json(v) for k, v in doc.items() if k != "ts"},
             },
             default=lambda o: None,
@@ -109,21 +90,26 @@ def encode_t2_body(t2: "T2Document") -> str:
     ).decode()
 
 
-def decode_t2_body(blob: Union[str,Dict[str,Any]]) -> Dict[str, Any]:
-    doc = json.loads(base64.b64decode(blob.encode()).decode()) if isinstance(blob, str) else blob
+def decode_t2_body(blob: str | dict[str, Any]) -> dict[str, Any]:
+    doc = (
+        json.loads(base64.b64decode(blob.encode()).decode())
+        if isinstance(blob, str)
+        else blob
+    )
     return {"ts": int(datetime.fromisoformat(doc.pop("timestamp")).timestamp()), **doc}
 
 
-def get_t2_result(t2: "T2Document") -> Union[Tuple[None, None], Tuple[datetime, Dict[str, Any]]]:
-    assert t2["body"] is not None
-    for record in reversed(t2["body"]):
-        if record.get("status", 0) == 0:
+def get_t2_result(
+    t2: "T2DocView",
+) -> tuple[None, None] | tuple[datetime, dict[str, Any]]:
+    assert t2.body is not None
+    for meta, record in zip(reversed(t2.meta), reversed(t2.body)):
+        if meta.get("code", DocumentCode.OK) == DocumentCode.OK:
             break
     else:
         return None, None
-    return datetime.fromtimestamp(record["ts"]), (
-        r[-1] if isinstance(r := record["result"], list) else r
-    )
+    assert isinstance(record, dict)
+    return datetime.fromtimestamp(meta["ts"]), record
 
 
 def render_thumbnail(cutout_data: bytes) -> str:
@@ -160,47 +146,34 @@ class SkyPortalAPIError(IOError):
     ...
 
 
-class UrlPathError(pydantic.errors.UrlError):
-    code = "url.path"
-    msg_template = "URL path must be empty, not {path!r}"
-
-
-class BaseHttpUrl(AnyHttpUrl):
-    """An http(s) URL with path unset"""
-
-    @classmethod
-    def validate(
-        cls, value: Any, field: "ModelField", config: "BaseConfig"
-    ) -> "AnyUrl":
-        url = super().validate(value, field, config)
-        if url.path is not None:
-            raise UrlPathError(path=url.path)
-        return url
-
-    # @classmethod
-    # def validate_parts(cls, parts: Dict[str, str]) -> Dict[str, str]:
-    #     parts["path"] = None
-    #     return super().validate_parts(parts)
-
-
-class SkyPortalClient(AmpelBaseModel):
+class SkyPortalClient(AmpelUnit):
 
     #: Base URL of SkyPortal server
-    base_url: BaseHttpUrl
+    base_url: str
     #: API token
-    token: Secret[str]
+    token: NamedSecret[str]
     #: Maximum number of in-flight requests
     max_parallel_connections: int = 1
 
+    @classmethod
+    def validate(cls, value: dict) -> Any:
+        value = super().validate(value)
+        url = urlparse(value["base_url"])
+        if url.scheme not in ("http", "https"):
+            raise ValueError("base_url must be http(s)")
+        if value["base_url"].endswith("/"):
+            raise ValueError("base_url may not have a path set")
+        return value
+
     def __init__(self, **kwargs) -> None:
         super().__init__(**kwargs)
 
         self._request_kwargs = {
             "headers": {"Authorization": f"token {self.token.get()}"}
         }
-        self._ids: Dict[str, Dict[str, int]] = {}
-        self._session: Optional[aiohttp.ClientSession] = None
-        self._semaphore: Optional[asyncio.Semaphore] = None
+        self._ids: dict[str, dict[str, int]] = {}
+        self._session: None | aiohttp.ClientSession = None
+        self._semaphore: None | asyncio.Semaphore = None
 
     @asynccontextmanager
     async def session(self, limit_per_host=0):
@@ -220,7 +193,7 @@ class SkyPortalClient(AmpelBaseModel):
         endpoint: str,
         raise_exc: bool,
         _decode_json: None,
-        **kwargs: Dict[str, Any],
+        **kwargs: dict[str, Any],
     ) -> aiohttp.ClientResponse:
         ...
 
@@ -231,8 +204,8 @@ class SkyPortalClient(AmpelBaseModel):
         endpoint: str,
         raise_exc: bool,
         _decode_json: bool,
-        **kwargs: Dict[str, Any],
-    ) -> Dict[str, Any]:
+        **kwargs: dict[str, Any],
+    ) -> dict[str, Any]:
         ...
 
     @backoff.on_exception(
@@ -250,9 +223,9 @@ class SkyPortalClient(AmpelBaseModel):
         verb: str,
         endpoint: str,
         raise_exc: bool = True,
-        _decode_json: Optional[bool] = True,
-        **kwargs: Dict[str, Any],
-    ) -> Union[aiohttp.ClientResponse, Dict[str, Any]]:
+        _decode_json: None | bool = True,
+        **kwargs: dict[str, Any],
+    ) -> aiohttp.ClientResponse | dict[str, Any]:
         if self._session is None or self._semaphore is None:
             raise ValueError(
                 "call operations within an `async with self.session()` block"
@@ -266,7 +239,7 @@ class SkyPortalClient(AmpelBaseModel):
             with stat_http_time.labels(*labels).time(), stat_http_errors.labels(
                 *labels
             ).count_exceptions(
-                (
+                ( # type: ignore[arg-type]
                     aiohttp.ClientResponseError,
                     aiohttp.ClientConnectionError,
                     asyncio.TimeoutError,
@@ -284,7 +257,10 @@ class SkyPortalClient(AmpelBaseModel):
                         payload = await response.json(content_type=None)
                         if raise_exc:
                             # only check status if endpoint knows it was returning JSON
-                            if response.content_type == "application/json" and payload["status"] != "success":
+                            if (
+                                response.content_type == "application/json"
+                                and payload["status"] != "success"
+                            ):
                                 raise SkyPortalAPIError(payload["message"], url, kwargs)
                             # otherwise, believe status code
                             else:
@@ -322,13 +298,13 @@ class SkyPortalClient(AmpelBaseModel):
             pass
         raise KeyError(f"No {endpoint} named {name}")
 
-    async def get(self, endpoint: str, **kwargs) -> Dict[str, Any]:
+    async def get(self, endpoint: str, **kwargs) -> dict[str, Any]:
         return await self.request("GET", endpoint, **kwargs)
 
-    async def post(self, endpoint: str, **kwargs) -> Dict[str, Any]:
+    async def post(self, endpoint: str, **kwargs) -> dict[str, Any]:
         return await self.request("POST", endpoint, **kwargs)
 
-    async def put(self, endpoint: str, **kwargs) -> Dict[str, Any]:
+    async def put(self, endpoint: str, **kwargs) -> dict[str, Any]:
         return await self.request("PUT", endpoint, **kwargs)
 
     async def head(self, endpoint: str, **kwargs) -> aiohttp.ClientResponse:
@@ -341,7 +317,7 @@ class FilterGroupProvisioner(SkyPortalClient):
     """
 
     #: mapping from ampel stream name to Fritz stream name
-    stream_names: Dict[str, str] = {
+    stream_names: dict[str, str] = {
         "ztf_uw_public": "ZTF Public",
         "ztf_uw_private": "ZTF Public+Partnership",
         "ztf_uw_caltech": "ZTF Public+Partnership+Caltech",
@@ -364,7 +340,7 @@ class FilterGroupProvisioner(SkyPortalClient):
         return await self.get_by_name("filters", name)
 
     async def create_filters(
-        self, config: "AmpelConfig", group: str, stream: Optional[str] = None
+        self, config: "AmpelConfig", group: str, stream: None | str = None
     ) -> None:
         """
         Create a dummy SkyPortal filter for each Ampel filter
@@ -454,19 +430,19 @@ async def provision_seed_data(client: SkyPortalClient):
 
 class PostReport(TypedDict):
     new: bool  #: is this a new source?
-    candidates: List[int]  #: new candidates created
-    save_error: Optional[str]  #: error raised while saving source
+    candidates: list[int]  #: new candidates created
+    save_error: None | str  #: error raised while saving source
     photometry_count: int  #: size of posted photometry
-    photometry_error: Optional[str]  #: error raised while posting photometry
+    photometry_error: None | str  #: error raised while posting photometry
     thumbnail_count: int  #: number of thumbnails posted
     comments: int  #: number of comments posted
-    comment_errors: List[str]  #: errors raised while posting comments
+    comment_errors: list[str]  #: errors raised while posting comments
     dt: float  #: elapsed time in seconds
 
 
 class BaseSkyPortalPublisher(SkyPortalClient):
 
-    logger: LoggerProtocol
+    logger: Traceless[LoggerProtocol]
 
     def __init__(self, **kwargs):
         if "logger" not in kwargs:
@@ -475,7 +451,7 @@ class BaseSkyPortalPublisher(SkyPortalClient):
 
     def _transform_datapoints(
         self, dps: Sequence["DataPoint"], after=-float("inf")
-    ) -> Generator[Dict[str, Any], None, None]:
+    ) -> Generator[dict[str, Any], None, None]:
         for dp in dps:
             body = dp["body"]
             if body["jd"] <= after:
@@ -483,7 +459,7 @@ class BaseSkyPortalPublisher(SkyPortalClient):
             if "SUPERSEDED" in dp["tag"]:
                 continue
             base = {
-                "_id": dp["_id"],
+                "id": dp["id"],
                 "filter": ZTF_FILTERS[body["fid"]],
                 "mjd": body["jd"] - 2400000.5,
                 "limiting_mag": body["diffmaglim"],
@@ -501,14 +477,14 @@ class BaseSkyPortalPublisher(SkyPortalClient):
 
     def make_photometry(
         self, datapoints: Sequence["DataPoint"], after=-float("inf")
-    ) -> Dict[str, Any]:
+    ) -> dict[str, Any]:
         content = defaultdict(list)
         for doc in self._transform_datapoints(datapoints, after):
             for k, v in doc.items():
                 content[k].append(v)
         return dict(content)
 
-    async def _find_instrument(self, tags: Sequence[Union[int, str]]) -> int:
+    async def _find_instrument(self, tags: Sequence[int | str]) -> int:
         for tag in tags:
             try:
                 return await self.get_by_name("instrument", tag)
@@ -519,27 +495,27 @@ class BaseSkyPortalPublisher(SkyPortalClient):
     async def post_t2_annotations(
         self,
         name: str,
-        t2_docs: Iterable["T2Document"],
-        object_record: Optional[Dict[str, Any]],
+        t2_views: Iterable["T2DocView"],
+        object_record: None | dict[str, Any],
         ret: PostReport,
     ):
         previous_annotations = object_record["annotations"] if object_record else []
-        for t2 in t2_docs:
+        for t2 in t2_views:
             last_modified, result = get_t2_result(t2)
             if result is None or last_modified is None:
                 continue
             # find associated annotation
             for annotation in previous_annotations:
-                if ":".split(annotation["origin"])[-1] == t2["unit"]:
+                if ":".split(annotation["origin"])[-1] == t2.unit:
                     break
             else:
                 # post new annotation
-                self.logger.debug(f"posting {t2['unit']}")
+                self.logger.debug(f"posting {t2.unit}")
                 try:
                     await self.post(
                         f"sources/{name}/annotation",
                         json={
-                            "origin": f"ampel:{t2['unit']}",
+                            "origin": f"ampel:{t2.unit}",
                             "data": flatten_dict(result, ":"),
                         },
                     )
@@ -549,14 +525,14 @@ class BaseSkyPortalPublisher(SkyPortalClient):
                 continue
             # update previous annotation
             if last_modified > datetime.fromisoformat(annotation["updated"]):
-                self.logger.debug(f"updating {t2['unit']}")
+                self.logger.debug(f"updating {t2.unit}")
                 try:
                     await self.put(
                         f"sources/{name}/annotation/{annotation['id']}",
                         json={
                             "obj_id": name,
                             "author_id": annotation["author_id"],
-                            "origin": f"ampel:{t2['unit']}",
+                            "origin": f"ampel:{t2.unit}",
                             "data": flatten_dict(result, ":"),
                         },
                     )
@@ -564,33 +540,33 @@ class BaseSkyPortalPublisher(SkyPortalClient):
                 except SkyPortalAPIError as exc:
                     ret["comment_errors"].append(exc.args[0])
             else:
-                self.logger.debug(f"{t2['unit']} exists and is current")
+                self.logger.debug(f"{t2.unit} exists and is current")
 
     async def post_t2_comments(
         self,
         name: str,
-        t2_docs: Iterable["T2Document"],
-        object_record: Optional[Dict[str, Any]],
+        t2_views: Iterable["T2DocView"],
+        object_record: None | dict[str, Any],
         ret: PostReport,
     ):
         previous_comments = object_record["comments"] if object_record else []
 
-        for t2 in t2_docs:
+        for t2 in t2_views:
             # find associated comment
             for comment in previous_comments:
-                if comment["text"] == t2["unit"]:
+                if comment["text"] == t2.unit:
                     break
             else:
                 # post new comment
-                self.logger.debug(f"posting {t2['unit']}")
+                self.logger.debug(f"posting {t2.unit}")
                 try:
                     await self.post(
                         f"sources/{name}/comment",
                         json={
-                            "text": t2["unit"],
+                            "text": t2.unit,
                             "attachment": {
                                 "body": encode_t2_body(t2),
-                                "name": f"{name}-{t2['unit']}.json",
+                                "name": f"{name}-{t2.unit}.json",
                             },
                         },
                     )
@@ -604,10 +580,8 @@ class BaseSkyPortalPublisher(SkyPortalClient):
                     f"sources/{name}/comments/{comment['id']}/attachment",
                 )
             )
-            if (t2["body"] is not None) and (
-                t2["body"][-1]["ts"] > previous_body["ts"]
-            ):
-                self.logger.debug(f"updating {t2['unit']}")
+            if (t2.body is not None) and (t2.meta[-1]["ts"] > previous_body["ts"]):
+                self.logger.debug(f"updating {t2.unit}")
                 try:
                     await self.put(
                         f"sources/{name}/comment/{comment['id']}",
@@ -622,14 +596,14 @@ class BaseSkyPortalPublisher(SkyPortalClient):
                 except SkyPortalAPIError as exc:
                     ret["comment_errors"].append(exc.args[0])
             else:
-                self.logger.debug(f"{t2['unit']} exists and is current")
+                self.logger.debug(f"{t2.unit} exists and is current")
 
     async def post_candidate(
         self,
         view: "TransientView",
-        filters: Optional[List[str]] = None,
-        groups: Optional[List[str]] = None,
-        instrument: Optional[str] = None,
+        filters: None | list[str] = None,
+        groups: None | list[str] = None,
+        instrument: None | str = None,
         post_photometry: bool = True,
         annotate: bool = False,
     ) -> PostReport:
@@ -694,7 +668,7 @@ class BaseSkyPortalPublisher(SkyPortalClient):
         assert view.t0 is not None
         dps = sorted(view.t0, key=lambda pp: pp["body"]["jd"])
         # detections, in time order
-        pps = [pp for pp in dps if pp["_id"] > 0]
+        pps = [pp for pp in dps if pp["id"] > 0]
 
         # post transient
         candidate = {
@@ -719,7 +693,7 @@ class BaseSkyPortalPublisher(SkyPortalClient):
 
         if (
             response := await self.get(
-                f"candidates/{name}", params={"includeComments": 1}, raise_exc=False
+                f"candidates/{name}", params={"includeComments": 1, "includeAlerts": 1}, raise_exc=False
             )
         )["status"] == "success":
             # Only update filters, not the candidate itself
@@ -731,8 +705,8 @@ class BaseSkyPortalPublisher(SkyPortalClient):
         # finding the time and alert id at which each filter passed for the
         # first time. If the candidate has not yet been posted, do so for the
         # first alert that passes a filter.
-        for jentry in view.get_journal_entries(tier=0) or []:
-            if jentry["extra"] is None or jentry["extra"].get("ac", False):
+        for jentry in view.get_journal_entries(tier=0):
+            if jentry.get("extra", {}).get("ac", False):
                 continue
             # no more filters left to update
             if not new_filters:
@@ -758,7 +732,7 @@ class BaseSkyPortalPublisher(SkyPortalClient):
                     json={
                         "id": name,
                         "filter_ids": fids,
-                        "passing_alert_id": jentry["extra"]["alert"],
+                        "passing_alert_id": jentry["alert"], # type: ignore[typeddict-item]
                         "passed_at": datetime.fromtimestamp(jentry["ts"]).isoformat(),
                         **candidate,
                     },
@@ -795,7 +769,7 @@ class BaseSkyPortalPublisher(SkyPortalClient):
                 "instrument_id": instrument_id,
                 **self.make_photometry(dps),
             }
-            datapoint_ids = photometry.pop("_id")
+            datapoint_ids = photometry.pop("id")
             try:
                 photometry_response = await self.put("photometry", json=photometry)
                 photometry_ids = photometry_response["data"]["ids"]
@@ -807,7 +781,7 @@ class BaseSkyPortalPublisher(SkyPortalClient):
         # SkyPortal only supports one of thumbnail per object and type
         # ('new', 'ref', 'sub', 'sdss', 'dr8', 'new_gz', 'ref_gz', 'sub_gz')
         # Post one of each type only if they do not yet exist.
-        existing_cutouts: Set[str] = (
+        existing_cutouts: set[str] = (
             {t["type"] for t in response["data"]["thumbnails"]}
             if response["status"] == "success"
             else set()
@@ -831,13 +805,16 @@ class BaseSkyPortalPublisher(SkyPortalClient):
                 ret["thumbnail_count"] += 1
 
         # represent latest T2 results as a comments
-        latest_t2: Dict[str, "T2Document"] = {}
+        latest_t2: dict[str, "T2DocView"] = {}
         for t2 in view.t2 or []:
-            if t2["status"] != T2RunState.COMPLETED or not t2["body"]:
+            if t2.code != DocumentCode.OK or not t2.body:
                 continue
-            assert isinstance(t2["unit"], str)
-            if t2["unit"] not in latest_t2 or latest_t2[t2["unit"]]["_id"] < t2["_id"]:
-                latest_t2[t2["unit"]] = t2
+            assert isinstance(t2.unit, str)
+            if (
+                t2.unit not in latest_t2
+                or latest_t2[t2.unit].meta[-1]["ts"] < t2.meta[-1]["ts"]
+            ):
+                latest_t2[t2.unit] = t2
 
         if annotate:
             await self.post_t2_annotations(
