@@ -1,23 +1,23 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# File              : Ampel-ZTF/ampel/ztf/t3/skyportal/SkyPortalPublisher.py
-# Author            : Jakob van Santen <jakob.van.santen@desy.de>
-# Date              : 16.09.2020
-# Last Modified Date: 16.09.2020
-# Last Modified By  : Jakob van Santen <jakob.van.santen@desy.de>
+# File:                Ampel-ZTF/ampel/ztf/t3/skyportal/SkyPortalPublisher.py
+# Author:              Jakob van Santen <jakob.van.santen@desy.de>
+# Date:                16.09.2020
+# Last Modified Date:  16.09.2020
+# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>
 
-import asyncio
+import asyncio, nest_asyncio
 from functools import partial
-from typing import Dict, List, Optional, Sequence, Tuple, TYPE_CHECKING
-
-import nest_asyncio
+from typing import TYPE_CHECKING
+from collections.abc import Generator
 
 from ampel.abstract.AbsPhotoT3Unit import AbsPhotoT3Unit
-from ampel.struct.JournalTweak import JournalTweak
-from ampel.type import StockId
+from ampel.struct.JournalAttributes import JournalAttributes
+from ampel.types import StockId
 from ampel.ztf.t3.skyportal.SkyPortalClient import BaseSkyPortalPublisher
 
 if TYPE_CHECKING:
+    from ampel.struct.T3Store import T3Store
     from ampel.view.TransientView import TransientView
     from ampel.content.JournalRecord import JournalRecord
 
@@ -25,19 +25,20 @@ if TYPE_CHECKING:
 class SkyPortalPublisher(BaseSkyPortalPublisher, AbsPhotoT3Unit):
 
     #: Save sources to these groups
-    groups: Optional[List[str]] = None
-    filters: Optional[List[str]] = None
+    groups: None | list[str] = None
+    filters: None | list[str] = None
     #: Post T2 results as annotations instead of comments
     annotate: bool = False
     #: Explicitly post photometry for each stock. If False, rely on some backend
     #: service (like Kowalski on Fritz) to fill in photometry for sources.
     include_photometry: bool = True
 
-    process_name: Optional[str] = None
+    process_name: None | str = None
 
-    def add(
-        self, tviews: Sequence["TransientView"]
-    ) -> Optional[Dict[StockId, JournalTweak]]:
+    def process(self,
+        tviews: Generator["TransientView", JournalAttributes, None],
+        t3s: 'None | T3Store' = None
+    ) -> None:
         """Pass each view to :meth:`post_candidate`."""
         # Patch event loop to be reentrant if it is already running, e.g.
         # within a notebook
@@ -47,7 +48,7 @@ class SkyPortalPublisher(BaseSkyPortalPublisher, AbsPhotoT3Unit):
         except RuntimeError:
             # second call raises: RuntimeError: There is no current event loop in thread 'MainThread'.
             ...
-        return asyncio.run(self.post_candidates(tviews))
+        asyncio.run(self.post_candidates(tviews))
 
     def _filter_journal_entries(self, jentry: "JournalRecord", after: float):
         """Select journal entries from SkyPortalPublisher newer than last update"""
@@ -77,35 +78,27 @@ class SkyPortalPublisher(BaseSkyPortalPublisher, AbsPhotoT3Unit):
         )
 
     async def post_candidates(
-        self, tviews: Sequence["TransientView"]
-    ) -> Dict[StockId, JournalTweak]:
+        self, tviews: Generator["TransientView", JournalAttributes, None]
+    ) -> None:
         """Pass each view to :meth:`post_candidate`."""
         async with self.session(limit_per_host=self.max_parallel_connections):
-            return dict(
-                await asyncio.gather(
-                    *[
-                        self.post_view(view)
-                        for view in tviews
-                        if self.requires_update(view)
-                    ],
-                )
+            await asyncio.gather(
+                *[
+                    self.post_view(view)
+                    for view in tviews
+                    if self.requires_update(view)
+                ],
             )
 
-    async def post_view(self, view: "TransientView") -> Tuple[StockId, JournalTweak]:
-        return (
-            view.id,
-            JournalTweak(
-                extra=dict(
-                    await self.post_candidate(
-                        view,
-                        groups=self.groups,
-                        filters=self.filters,
-                        annotate=self.annotate,
-                        post_photometry=self.include_photometry,
-                    )
+    async def post_view(self, view: "TransientView") -> tuple[StockId, JournalAttributes]:
+        return view.id, JournalAttributes(
+            extra=dict(
+                await self.post_candidate(
+                    view,
+                    groups=self.groups,
+                    filters=self.filters,
+                    annotate=self.annotate,
+                    post_photometry=self.include_photometry,
                 )
-            ),
+            )
         )
-
-    def done(self) -> None:
-        ...
